@@ -1,19 +1,15 @@
 /* Copyright (C) 2013-2016, The Regents of The University of Michigan.
 All rights reserved.
-
 This software was developed in the APRIL Robotics Lab under the
 direction of Edwin Olson, ebolson@umich.edu. This software may be
 available under alternative licensing terms; contact the address above.
-
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
-
 1. Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
-
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
@@ -24,7 +20,6 @@ LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-
 The views and conclusions contained in the software and documentation are those
 of the authors and should not be interpreted as representing official policies,
 either expressed or implied, of the Regents of The University of Michigan.
@@ -325,23 +320,26 @@ void image_u8_convolve_2D(image_u8_t *im, const uint8_t *k, int ksz)
 
     for (int y = 0; y < im->height; y++) {
 
-        uint8_t x[im->stride];
+        uint8_t *x = malloc(sizeof(uint8_t)*im->stride);
         memcpy(x, &im->buf[y*im->stride], im->stride);
 
         convolve(x, &im->buf[y*im->stride], im->width, k, ksz);
+        free(x);
     }
 
     for (int x = 0; x < im->width; x++) {
-        uint8_t xb[im->height];
-        uint8_t yb[im->height];
+        uint8_t *xb = malloc(sizeof(uint8_t)*im->height);
+        uint8_t *yb = malloc(sizeof(uint8_t)*im->height);
 
         for (int y = 0; y < im->height; y++)
             xb[y] = im->buf[y*im->stride + x];
 
         convolve(xb, yb, im->height, k, ksz);
+        free(xb);
 
         for (int y = 0; y < im->height; y++)
             im->buf[y*im->stride + x] = yb[y];
+        free(yb);
     }
 }
 
@@ -353,7 +351,7 @@ void image_u8_gaussian_blur(image_u8_t *im, double sigma, int ksz)
     assert((ksz & 1) == 1); // ksz must be odd.
 
     // build the kernel.
-    double dk[ksz];
+    double *dk = malloc(sizeof(double)*ksz);
 
     // for kernel of length 5:
     // dk[0] = f(-2), dk[1] = f(-1), dk[2] = f(0), dk[3] = f(1), dk[4] = f(2)
@@ -371,7 +369,7 @@ void image_u8_gaussian_blur(image_u8_t *im, double sigma, int ksz)
     for (int i = 0; i < ksz; i++)
         dk[i] /= acc;
 
-    uint8_t k[ksz];
+    uint8_t *k = malloc(sizeof(uint8_t)*ksz);
     for (int i = 0; i < ksz; i++)
         k[i] = dk[i]*255;
 
@@ -379,8 +377,10 @@ void image_u8_gaussian_blur(image_u8_t *im, double sigma, int ksz)
         for (int i = 0; i < ksz; i++)
             printf("%d %15f %5d\n", i, dk[i], k[i]);
     }
+    free(dk);
 
     image_u8_convolve_2D(im, k, ksz);
+    free(k);
 }
 
 image_u8_t *image_u8_rotate(const image_u8_t *in, double rad, uint8_t pad)
@@ -433,93 +433,6 @@ image_u8_t *image_u8_rotate(const image_u8_t *in, double rad, uint8_t pad)
     return out;
 }
 
-#ifdef __ARM_NEON__
-#include <arm_neon.h>
-
-void neon_decimate2(uint8_t * __restrict dest, int destwidth, int destheight, int deststride,
-                    uint8_t * __restrict src, int srcwidth, int srcheight, int srcstride)
-{
-    for (int y = 0; y < destheight; y++) {
-        for (int x = 0; x < destwidth; x+=8) {
-            uint8x16x2_t row0 = vld2q_u8(src + 2*x);
-            uint8x16x2_t row1 = vld2q_u8(src + 2*x + srcstride);
-            uint8x16_t sum0 = vhaddq_u8(row0.val[0], row1.val[1]);
-            uint8x16_t sum1 = vhaddq_u8(row1.val[0], row0.val[1]);
-            uint8x16_t sum = vhaddq_u8(sum0, sum1);
-            vst1q_u8(dest + x, sum);
-        }
-        src += 2*srcstride;
-        dest += deststride;
-    }
-}
-
-void neon_decimate3(uint8_t * __restrict dest, int destwidth, int destheight, int deststride,
-                    uint8_t * __restrict src, int srcwidth, int srcheight, int srcstride)
-{
-    for (int y = 0; y < destheight; y++) {
-        for (int x = 0; x < destwidth; x+=8) {
-            uint8x16x3_t row0 = vld3q_u8(src + 3*x);
-            uint8x16x3_t row1 = vld3q_u8(src + 3*x + srcstride);
-            uint8x16x3_t row2 = vld3q_u8(src + 3*x + 2*srcstride);
-
-            uint8x16_t sum0 = vhaddq_u8(row0.val[0], row0.val[1]);
-            uint8x16_t sum1 = vhaddq_u8(row0.val[2], row1.val[0]);
-            uint8x16_t sum2 = vhaddq_u8(row1.val[1], row1.val[2]);
-            uint8x16_t sum3 = vhaddq_u8(row2.val[0], row2.val[1]);
-
-            uint8x16_t suma = vhaddq_u8(sum0, sum1);
-            uint8x16_t sumb = vhaddq_u8(sum2, sum3);
-            uint8x16_t sum = vhaddq_u8(suma, sumb);
-
-            vst1q_u8(dest + x, sum);
-        }
-        src += 3*srcstride;
-        dest += deststride;
-    }
-}
-
-void neon_decimate4(uint8_t * __restrict dest, int destwidth, int destheight, int deststride,
-                    uint8_t * __restrict src, int srcwidth, int srcheight, int srcstride)
-{
-    for (int y = 0; y < destheight; y++) {
-        for (int x = 0; x < destwidth; x+=8) {
-            uint8x16x4_t row0 = vld4q_u8(src + 4*x);
-            uint8x16x4_t row1 = vld4q_u8(src + 4*x + srcstride);
-            uint8x16x4_t row2 = vld4q_u8(src + 4*x + 2*srcstride);
-            uint8x16x4_t row3 = vld4q_u8(src + 4*x + 3*srcstride);
-
-            uint8x16_t sum0, sum1;
-
-            sum0 = vhaddq_u8(row0.val[0], row0.val[3]);
-            sum1 = vhaddq_u8(row0.val[2], row0.val[1]);
-            uint8x16_t suma = vhaddq_u8(sum0, sum1);
-
-            sum0 = vhaddq_u8(row1.val[0], row1.val[3]);
-            sum1 = vhaddq_u8(row1.val[2], row1.val[1]);
-            uint8x16_t sumb = vhaddq_u8(sum0, sum1);
-
-            sum0 = vhaddq_u8(row2.val[0], row2.val[3]);
-            sum1 = vhaddq_u8(row2.val[2], row2.val[1]);
-            uint8x16_t sumc = vhaddq_u8(sum0, sum1);
-
-            sum0 = vhaddq_u8(row3.val[0], row3.val[3]);
-            sum1 = vhaddq_u8(row3.val[2], row3.val[1]);
-            uint8x16_t sumd = vhaddq_u8(sum0, sum1);
-
-            uint8x16_t sumx = vhaddq_u8(suma, sumd);
-            uint8x16_t sumy = vhaddq_u8(sumc, sumb);
-
-            uint8x16_t sum = vhaddq_u8(sumx, sumy);
-
-            vst1q_u8(dest + x, sum);
-        }
-        src += 4*srcstride;
-        dest += deststride;
-    }
-}
-
-#endif
-
 image_u8_t *image_u8_decimate(image_u8_t *im, float ffactor)
 {
     int width = im->width, height = im->height;
@@ -572,89 +485,18 @@ image_u8_t *image_u8_decimate(image_u8_t *im, float ffactor)
 
     int factor = (int) ffactor;
 
-    int swidth = width / factor, sheight = height / factor;
-
+    int swidth = 1 + (width - 1)/factor;
+    int sheight = 1 + (height - 1)/factor;
     image_u8_t *decim = image_u8_create(swidth, sheight);
-
-#ifdef __ARM_NEON__
-    if (factor == 2) {
-        neon_decimate2(decim->buf, decim->width, decim->height, decim->stride,
-                       im->buf, im->width, im->height, im->stride);
-        return decim;
-    } else if (factor == 3) {
-        neon_decimate3(decim->buf, decim->width, decim->height, decim->stride,
-                       im->buf, im->width, im->height, im->stride);
-        return decim;
-    } else if (factor == 4) {
-        neon_decimate4(decim->buf, decim->width, decim->height, decim->stride,
-                       im->buf, im->width, im->height, im->stride);
-        return decim;
+    int sy = 0;
+    for (int y = 0; y < height; y += factor) {
+        int sx = 0;
+        for (int x = 0; x < width; x += factor) {
+            decim->buf[sy*decim->stride + sx] = im->buf[y*im->stride + x];
+            sx++;
+        }
+        sy++;
     }
-#endif
-
-    if (factor == 2) {
-        for (int sy = 0; sy < sheight; sy++) {
-            int sidx = sy * decim->stride;
-            int idx = (sy*2)*im->stride;
-
-            for (int sx = 0; sx < swidth; sx++) {
-                uint32_t v = im->buf[idx] + im->buf[idx+1] +
-                    im->buf[idx+im->stride] + im->buf[idx+im->stride + 1];
-                decim->buf[sidx] = (v>>2);
-                idx+=2;
-                sidx++;
-            }
-        }
-    } else if (factor == 3) {
-        for (int sy = 0; sy < sheight; sy++) {
-            int sidx = sy * decim->stride;
-            int idx = (sy*3)*im->stride;
-
-            for (int sx = 0; sx < swidth; sx++) {
-                uint32_t v = im->buf[idx] + im->buf[idx+1] + im->buf[idx+2] +
-                    im->buf[idx+im->stride] + im->buf[idx+im->stride + 1] + im->buf[idx+im->stride + 2] +
-                    im->buf[idx+2*im->stride] + im->buf[idx+2*im->stride + 1];
-                // + im->buf[idx+2*im->stride + 1];
-                // deliberately omit lower right corner so there are exactly 8 samples...
-                decim->buf[sidx] = (v>>3);
-                idx+=3;
-                sidx++;
-            }
-        }
-    } else if (factor == 4) {
-        for (int sy = 0; sy < sheight; sy++) {
-            int sidx = sy * decim->stride;
-            int idx = (sy*4)*im->stride;
-
-            for (int sx = 0; sx < swidth; sx++) {
-                uint32_t v = im->buf[idx] + im->buf[idx+1] + im->buf[idx+2] + im->buf[idx+3] +
-                    im->buf[idx+im->stride] + im->buf[idx+im->stride + 1] + im->buf[idx+im->stride + 1] + im->buf[idx+im->stride + 2] +
-                    im->buf[idx+2*im->stride] + im->buf[idx+2*im->stride + 1] + im->buf[idx+2*im->stride + 2] + im->buf[idx+2*im->stride + 3];
-
-                decim->buf[sidx] = (v>>4);
-                idx+=4;
-                sidx++;
-            }
-        }
-    } else {
-        // XXX this isn't a very good decimation code.
-        uint32_t row[swidth];
-
-        for (int y = 0; y < height; y+= factor) {
-            memset(row, 0, sizeof(row));
-
-            for (int dy = 0; dy < factor; dy++) {
-                for (int x = 0; x < width; x++) {
-                    row[x/factor] += im->buf[(y+dy)*im->stride + x];
-                }
-            }
-
-            for (int x = 0; x < swidth; x++)
-                decim->buf[(y/factor)*decim->stride + x] = row[x] / sq(factor);
-
-        }
-    }
-
     return decim;
 }
 
