@@ -1,19 +1,15 @@
 /* Copyright (C) 2013-2016, The Regents of The University of Michigan.
 All rights reserved.
-
 This software was developed in the APRIL Robotics Lab under the
 direction of Edwin Olson, ebolson@umich.edu. This software may be
 available under alternative licensing terms; contact the address above.
-
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
-
 1. Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
-
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
@@ -24,7 +20,6 @@ LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-
 The views and conclusions contained in the software and documentation are those
 of the authors and should not be interpreted as representing official policies,
 either expressed or implied, of the Regents of The University of Michigan.
@@ -816,7 +811,7 @@ matd_t *matd_op(const char *expr, ...)
     va_list ap;
     va_start(ap, expr);
 
-    matd_t *args[nargs];
+    matd_t **args = malloc(sizeof(matd_t*)*nargs);
     for (int i = 0; i < nargs; i++) {
         args[i] = va_arg(ap, matd_t*);
         // XXX: sanity check argument; emit warning/error if args[i]
@@ -829,10 +824,12 @@ matd_t *matd_op(const char *expr, ...)
     int argpos = 0;
     int garbpos = 0;
 
-    matd_t *garb[2*exprlen]; // can't create more than 2 new result per character
-                             // one result, and possibly one argument to free
+    // can't create more than 2 new result per character
+    // one result, and possibly one argument to free
+    matd_t **garb = malloc(sizeof(matd_t*)*2*exprlen);
 
     matd_t *res = matd_op_recurse(expr, &pos, NULL, args, &argpos, garb, &garbpos, 0);
+    free(args);
 
     // 'res' may need to be freed as part of garbage collection (i.e. expr = "F")
     matd_t *res_copy = (res ? matd_copy(res) : NULL);
@@ -840,6 +837,7 @@ matd_t *matd_op(const char *expr, ...)
     for (int i = 0; i < garbpos; i++) {
         matd_destroy(garb[i]);
     }
+    free(garb);
 
     return res_copy;
 }
@@ -1026,7 +1024,7 @@ static matd_svd_t matd_svd_tall(matd_t *A, int flags)
             //
             int vlen = A->nrows - hhidx;
 
-            double v[vlen];
+            double *v = malloc(sizeof(double)*vlen);
 
             double mag2 = 0;
             for (int i = 0; i < vlen; i++) {
@@ -1046,8 +1044,10 @@ static matd_svd_t matd_svd_tall(matd_t *A, int flags)
             double mag = sqrt(mag2);
 
             // this case arises with matrices of all zeros, for example.
-            if (mag == 0)
+            if (mag == 0) {
+                free(v);
                 continue;
+            }
 
             for (int i = 0; i < vlen; i++)
                 v[i] /= mag;
@@ -1078,12 +1078,14 @@ static matd_svd_t matd_svd_tall(matd_t *A, int flags)
                 for (int j = 0; j < vlen; j++)
                     MATD_EL(B, hhidx+j, i) -= 2*dot*v[j];
             }
+
+            free(v);
         }
 
         if (hhidx+2 < A->ncols) {
             int vlen = A->ncols - hhidx - 1;
 
-            double v[vlen];
+            double *v = malloc(sizeof(double)*vlen);
 
             double mag2 = 0;
             for (int i = 0; i < vlen; i++) {
@@ -1103,8 +1105,10 @@ static matd_svd_t matd_svd_tall(matd_t *A, int flags)
             double mag = sqrt(mag2);
 
             // this case can occur when the vectors are already perpendicular
-            if (mag == 0)
+            if (mag == 0) {
+                free(v);
                 continue;
+            }
 
             for (int i = 0; i < vlen; i++)
                 v[i] /= mag;
@@ -1132,13 +1136,15 @@ static matd_svd_t matd_svd_tall(matd_t *A, int flags)
                 for (int j = 0; j < vlen; j++)
                     MATD_EL(B, i, hhidx+1+j) -= 2*dot*v[j];
             }
+
+            free(v);
         }
     }
 
-    // maxiters used to be smaller to prevent us from looping forever,
-    // but this doesn't seem to happen any more with our more stable
-    // svd22 implementation.
-    int maxiters = 1UL << 30;
+    // empirically, we find a roughly linear worst-case number of iterations
+    // as a function of rows*cols. maxiters ~= 1.5*nrows*ncols
+    // we're a bit conservative below.
+    int maxiters = 200 + 2 * A->nrows * A->ncols;
     assert(maxiters > 0); // reassure clang
     int iter;
 
@@ -1152,7 +1158,7 @@ static matd_svd_t matd_svd_tall(matd_t *A, int flags)
 
     // for each of the first B->ncols rows, which index has the
     // maximum absolute value? (used by method 1)
-    int maxrowidx[B->ncols];
+    int *maxrowidx = malloc(sizeof(int)*B->ncols);
     int lastmaxi, lastmaxj;
 
     if (find_max_method == 1) {
@@ -1366,6 +1372,8 @@ static matd_svd_t matd_svd_tall(matd_t *A, int flags)
         }
     }
 
+    free(maxrowidx);
+
     if (!(flags & MATD_SVD_NO_WARNINGS) && iter == maxiters) {
         printf("WARNING: maximum iters (maximum = %d, matrix %d x %d, max=%.15f)\n",
                iter, A->nrows, A->ncols, maxv);
@@ -1375,8 +1383,8 @@ static matd_svd_t matd_svd_tall(matd_t *A, int flags)
 
     // them all positive by flipping the corresponding columns of
     // U/LS.
-    int idxs[A->ncols];
-    double vals[A->ncols];
+    int *idxs = malloc(sizeof(int)*A->ncols);
+    double *vals = malloc(sizeof(double)*A->ncols);
     for (int i = 0; i < A->ncols; i++) {
         idxs[i] = i;
         vals[i] = MATD_EL(B, i, i);
@@ -1412,6 +1420,8 @@ static matd_svd_t matd_svd_tall(matd_t *A, int flags)
         MATD_EL(LP, idxs[i], i) = vals[i] < 0 ? -1 : 1;
         MATD_EL(RP, idxs[i], i) = 1; //vals[i] < 0 ? -1 : 1;
     }
+    free(idxs);
+    free(vals);
 
     // we've factored:
     // LP*(something)*RP'
@@ -1536,7 +1546,7 @@ matd_plu_t *matd_plu(const matd_t *a)
 
         // swap rows p and j?
         if (p != j) {
-            TYPE tmp[lu->ncols];
+            TYPE *tmp = malloc(sizeof(TYPE)*lu->ncols);
             memcpy(tmp, &MATD_EL(lu, p, 0), sizeof(TYPE) * lu->ncols);
             memcpy(&MATD_EL(lu, p, 0), &MATD_EL(lu, j, 0), sizeof(TYPE) * lu->ncols);
             memcpy(&MATD_EL(lu, j, 0), tmp, sizeof(TYPE) * lu->ncols);
@@ -1544,6 +1554,7 @@ matd_plu_t *matd_plu(const matd_t *a)
             piv[p] = piv[j];
             piv[j] = k;
             pivsign = -pivsign;
+            free(tmp);
         }
 
         double LUjj = MATD_EL(lu, j, j);
